/* REDC based modular arithmetic.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include <stdlib.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "mpm.h"
#include "mpm-impl.h"


/* The "R" from Montgomery's paper is msize*2^GMP_NUMB_BITS.  Residues are
   kept in the range 0 <= r < m.  There's two blocks of code, a "big" and a
   "small".

   In the small case, reduction is done with an mpn_addmul_1 for each limb
   to be zapped, the same as in mpz_powm.  For the big case a large inverse
   is pre-calculated and the block multiplies mpn_mullowhalf_n and
   mpn_mulhighhalf_n are used.  This lets Karatsuba etc in those functions
   bring the time down to sub-quadratic.  (There's no actual
   mpn_mulhighhalf_n yet though, it's just the full mpn_mul_n.)

   Enhancements:

   In the "small" case, a next step can be initiated as soon as the first
   two limbs of the mpn_addmul_1 have been done.  Perhaps some use of that
   could be made to keep a pipelined multiplier busy, ie. there wouldn't be
   a warm up and warm down phase in between addmul_1 steps.

   The usual sizes for RSA encryption are 512 or 1024 bit moduli.  That's
   only 8 or 16 limbs on a 64-bit system and could almost fit entirely in
   registers for the redc reduction step, or even for an entire mpr_pow.  */


/* change this to "#define TRACE(x) x" for diagnostics */
#define TRACE(x) 

#ifndef REDC_BIG_THRESHOLD
#define REDC_BIG_THRESHOLD  100
#endif

#ifndef REDC_SETZ_THRESHOLD
#define REDC_SETZ_THRESHOLD  0  /* always */
#endif


#ifndef HAVE_NATIVE_mpn_addmul_2
#define HAVE_NATIVE_mpn_addmul_2  0
#endif
#ifndef HAVE_NATIVE_mpn_addmul_3
#define HAVE_NATIVE_mpn_addmul_3  0
#endif
#ifndef HAVE_NATIVE_mpn_addmul_4
#define HAVE_NATIVE_mpn_addmul_4  0
#endif


#if HAVE_NATIVE_mpn_addmul_4
#define MSMALLINV_SIZE  4
#else
#if HAVE_NATIVE_mpn_addmul_3
#define MSMALLINV_SIZE  3
#else
#if HAVE_NATIVE_mpn_addmul_2
#define MSMALLINV_SIZE  2
#else
#define MSMALLINV_SIZE  1
#endif
#endif
#endif


#if HAVE_NATIVE_mpn_addmul_2    \
  || HAVE_NATIVE_mpn_addmul_3   \
  || HAVE_NATIVE_mpn_addmul_4

#define REDC_SMALL(rp, pp, m, mp, msize)   mpr_redc1__redc_small (rp, pp, m)
static void
mpr_redc1__redc_small (mp_ptr rp, mp_ptr pp, mpm_srcptr m)
{
  mp_srcptr  mp;
  mp_limb_t  c, c2, minv3, minv2, minv1, minv0;
  mp_size_t  msize, i;

  msize = MRSIZE(m);
  mp = MPTR(m);
  i = 0;

  if (HAVE_NATIVE_mpn_addmul_4)
    {
      ASSERT (MSMALLINV_SIZE >= 4);
      minv0 = MSMALLINV(m)[0];
      minv1 = MSMALLINV(m)[1];
      minv2 = MSMALLINV(m)[2];
      minv3 = MSMALLINV(m)[3];

      for ( ; i <= msize-4; i += 4)
        {
          mp_limb_t  q[4];
          MUL_444 (q[3], q[2], q[1], q[0],
                   pp[i+3], pp[i+2], pp[i+1], pp[i],
                   minv3, minv2, minv1, minv0);
          c = mpn_addmul_4 (pp+i, mp, msize, q);
          ASSERT (pp[i] == 0);
          ASSERT (pp[i+1] == 0);
          ASSERT (pp[i+2] == 0);
          ASSERT (pp[i+3] == 0);
          pp[i+2] = c;
        }
    }

  if (HAVE_NATIVE_mpn_addmul_3)
    {
      if (! HAVE_NATIVE_mpn_addmul_4)
        {
          ASSERT (MSMALLINV_SIZE >= 3);
          minv0 = MSMALLINV(m)[0];
          minv1 = MSMALLINV(m)[1];
          minv2 = MSMALLINV(m)[2];
        }

      for ( ; i <= msize-3; i += 3)
        {
          mp_limb_t  q[3];
          MUL_333 (q[2], q[1], q[0],
                   pp[i+2], pp[i+1], pp[i],
                   minv2, minv1, minv0);
          c = mpn_addmul_3 (pp+i, mp, msize, q);
          ASSERT (pp[i] == 0);
          ASSERT (pp[i+1] == 0);
          ASSERT (pp[i+2] == 0);
          pp[i+2] = c;

          /* an addmul_4 means at most 3 left */
          if (HAVE_NATIVE_mpn_addmul_4)
            goto carries;
        }
    }

  if (HAVE_NATIVE_mpn_addmul_2)
    {
      if (! (HAVE_NATIVE_mpn_addmul_4 || HAVE_NATIVE_mpn_addmul_3))
        {
          ASSERT (MSMALLINV_SIZE >= 2);
          minv0 = MSMALLINV(m)[0];
          minv1 = MSMALLINV(m)[1];
        }

      for ( ; i <= msize-2; i += 2)
        {
          mp_limb_t  q[2];
          MUL_222 (q[1],q[0], pp[i+1],pp[i], minv1,minv0);
          c = mpn_addmul_2 (pp+i, mp, msize, q);
          ASSERT (pp[i] == 0);
          ASSERT (pp[i+1] == 0);
          pp[i+1] = c;

          /* an addmul_3 means at most 2 left */
          if (HAVE_NATIVE_mpn_addmul_3)
            goto carries;
        }
    }

  if (! (HAVE_NATIVE_mpn_addmul_4
         || HAVE_NATIVE_mpn_addmul_3
         || HAVE_NATIVE_mpn_addmul_2))
    {
      minv0 = MSMALLINV(m)[0];
    }

  for ( ; i < msize; i++)
    {
      mp_limb_t  q;
      q = (pp[i] * minv0) & GMP_NUMB_MASK;
      TRACE (mpn_trace ("  q", &q, (mp_size_t) 1));
      c = mpn_addmul_1 (pp+i, mp, msize, q);
      TRACE (mpn_trace ("  sum", pp+i, msize));
      ASSERT (pp[i] == 0);
      pp[i] = c;

      /* an addmul_2 means at most 1 left */
      if (HAVE_NATIVE_mpn_addmul_2)
        goto carries;
    }

 carries:
  /* apply addmul carries */
  TRACE (mpn_trace ("  add", pp, msize);
         mpn_trace ("   to", pp+msize, msize));
  c = mpn_add_n (rp, pp+msize, pp, msize);
  TRACE (mpn_trace ("    c", &c, (mp_size_t) 1);
         mpn_trace ("   rp", rp, msize));

  /* subtract an m if r>=m, which should leave r<m */
  if (LIKELY (c == 0))
    MPN_CMP_LT_ACTION (rp, mp, msize, goto done);

  TRACE (printf ("  subtract\n"));
  c2 = mpn_sub_n (rp, rp, mp, msize);
  ASSERT (c2 == c);

 done:
  ASSERT (mpn_cmp (rp, mp, msize) < 0);
}

#else

#define REDC_SMALL(rp, pp, m, mp, msize)                        \
  do {                                                          \
    mp_ptr     __rp;                                            \
    mp_srcptr  __mp;                                            \
    mp_limb_t  __minv, __q, __c;                                \
    mp_size_t  __i;                                             \
                                                                \
    __mp = (mp);                                                \
    __minv = MSMALLINV(m)[0];                                   \
    for (__i = 0; __i < (msize); __i++)                         \
      {                                                         \
        __q = ((pp)[__i] * __minv) & GMP_NUMB_MASK;             \
        TRACE (mpn_trace ("  q", &__q, (mp_size_t) 1));         \
        __c = mpn_addmul_1 ((pp)+__i, __mp, msize, __q);        \
        TRACE (mpn_trace ("  sum", (pp)+__i, msize));           \
        ASSERT ((pp)[__i] == 0);                                \
        (pp)[__i] = __c;                                        \
                                                                \
        if (HAVE_NATIVE_mpn_addmul_2)                           \
          break;                                                \
      }                                                         \
                                                                \
    /* apply addmul carries */                                  \
    TRACE (mpn_trace ("  add", pp, msize);                      \
           mpn_trace ("   to", (pp)+(msize), msize));           \
    __rp = (rp);                                                \
    __c = mpn_add_n (__rp, (pp)+(msize), pp, msize);            \
    TRACE (mpn_trace ("    c", &__c, (mp_size_t) 1);            \
           mpn_trace ("   rp", __rp, msize));                   \
                                                                \
    /* not a loop, just avoid a label for the CMP action */     \
    for (;;)                                                    \
      {                                                         \
        mp_limb_t  __c2;                                        \
        /* subtract an m if r>=m, which should leave r<m */     \
        if (LIKELY (__c == 0))                                  \
          MPN_CMP_LT_ACTION (__rp, __mp, msize, break);         \
        TRACE (printf ("  subtract\n"));                        \
        __c2 = mpn_sub_n (__rp, __rp, __mp, msize);             \
        ASSERT (__c2 == __c);                                   \
        break;                                                  \
      }                                                         \
                                                                \
    ASSERT (mpn_cmp (__rp, __mp, msize) < 0);                   \
  } while (0)

#endif


#define REDC_BIG(rp, pp, tp, qp, mp, minv, msize)               \
  do {                                                          \
    mp_ptr     __rp;                                            \
    mp_srcptr  __mp;                                            \
    TRACE (mpn_trace ("  pp", pp, 2*msize));                    \
    mpn_mullowhalf_n (qp, pp, minv, msize);                     \
    TRACE (mpn_trace ("  qp", qp, msize));                      \
    ASSERT_CODE (mpn_mul_n (tp, qp, mp, msize));                \
    ASSERT (mpn_cmp (tp, pp, msize) == 0);                      \
    __mp = (mp);                                                \
    mpn_mulhighhalf_n (tp, qp, __mp, msize);                    \
    TRACE (mpn_trace ("  sub", tp, 2*msize));                   \
    __rp = (rp);                                                \
    if (mpn_sub_n (__rp, (pp)+(msize), (tp)+(msize), msize))    \
      {                                                         \
        TRACE (mpn_trace ("  diff", rp, msize));                \
        mpn_add_n (__rp, __rp, __mp, msize);                    \
      }                                                         \
    TRACE (mpn_trace ("  rp", rp, msize));                      \
  } while (0)


static void
mpr_redc__add_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;
  mp_limb_t  c;

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      mpr_init_set_ui (yr, y, m);
      mpr_add (w, x, yr, m);
      mpr_clear (yr, m);
      return;
    }

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);
  if (w != x)
    MPN_COPY (wp, xp, msize);
  c = mpn_addmul_1 (wp, RPTR(MRMOD(m)), msize, (mp_limb_t) y);
  mpr_mod__reduce_1 (c, wp, m);

  ASSERT_MPR_MOD (w, m);
}

static int
mpr_redc__equal_one (mpr_srcptr x, mpm_srcptr m)
{
  ASSERT_MPR_MOD (x, m);
  MPN_EQUAL_RETURN (RPTR(x), RPTR(MRMOD(m)), MRSIZE(m));
}

/* As a residue, y will be a value y*r-q*m, where r = R mod m and q is
   chosen to reduce to the range 0 to m-1.  q will be a single limb (so long
   as y is), and hence can be calculated as q=(y*r-x)/m mod 2^GMP_NUMB_BITS.
   A comparison is made between y*r and x+q*m, if they're the same then x
   and y are equal, or if not, not.  y*r is like an mpn_mul_1 and x+q*m is
   like an addmul_1, but they're calculated progressively from low to high,
   since for random inputs x!=y should be common and can be identified from
   just one or two limbs.

   Enhancement: Could do the same when y is 2 limbs (in the nails case),
   allowing for q to be 2 limbs too.  */

static int
mpr_redc__equal_ui (mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  ASSERT_MPR_MOD (x, m);

  if (y <= GMP_NUMB_MAX)
    {
      mp_srcptr  xp, mp, rp;
      mp_limb_t  q, yh, yl, yc, xh, xl, xc;
      mp_size_t  msize, i;

      TRACE (printf ("mpr_redc__equal_ui\n");
             mpr_trace ("  x   ", x, m);
             printf    ("  y   =%#lX\n", y);
             mpz_trace ("  m   ", MZ(m));
             mpr_trace ("  rmod", m->_mp_rmod, m);
             mpn_trace ("  msmallinv",
                        MSMALLINV(m), (mp_size_t) MSMALLINV_SIZE);
             );

      rp = RPTR(MRMOD(m));
      MUL_211 (yh,yl, y, rp[0]);

      xp = RPTR(x);
      q = ((xp[0] - yl) * MSMALLINV(m)[0]) & GMP_NUMB_MASK;
      TRACE (mpn_trace ("  q", &q, (mp_size_t) 1));

      mp = MPTR(m);
      MUL_211 (xh,xl, q, mp[0]);
      ADD_221 (xh,xl, xh,xl, xp[0]);

      TRACE (mpn_trace ("  xl", &xl, (mp_size_t) 1);
             mpn_trace ("  xh", &xh, (mp_size_t) 1);
             mpn_trace ("  yl", &yl, (mp_size_t) 1);
             mpn_trace ("  yh", &yh, (mp_size_t) 1));
      ASSERT (xl == yl);

      msize = MRSIZE(m);
      for (i = 1; i < msize; i++)
        {
          yc = yh;
          MUL_211 (yh,yl, y, rp[i]);
          ADD_221_NC (yh,yl, yh,yl, yc);

          xc = xh;
          MUL_211 (xh,xl, q, mp[i]);
          ADD_2211_NC (xh,xl, xh,xl, xc, xp[i]);

          TRACE (printf (" i=%ld\n", i);
                 mpn_trace ("  xl", &xl, (mp_size_t) 1);
                 mpn_trace ("  xh", &xh, (mp_size_t) 1);
                 mpn_trace ("  yl", &yl, (mp_size_t) 1);
                 mpn_trace ("  yh", &yh, (mp_size_t) 1));

          if (LIKELY (xl != yl))
            return 0;
        }

      if (xh != yh)
        return 0;
      return 1;
    }
  else
    {
      mpr_t      t;
      mp_size_t  msize;
      int        ret;
      TMP_DECL (marker);

      TMP_MARK (marker);
      msize = MRSIZE(m);
      t = (mpr_ptr) TMP_ALLOC_LIMBS (msize);
      mpr_mod__mul_ui (t, m->_mp_rmod, y, m);
      MPN_EQUAL (ret, RPTR(x), RPTR(t), msize);
      TMP_FREE (marker);
      return ret;
    }
}

static void
mpr_redc__get_z_small (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, wsize;
  mp_ptr     wp, pp;
  TMP_DECL (marker);

  TRACE (printf ("mpr_redc__get_z_small\n");
         mpn_trace ("  xp", RPTR(x), MRSIZE(m)));
  ASSERT_MPR_MOD (x, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  pp = TMP_ALLOC_LIMBS (2*msize);
  MPN_COPY_ZERO_EXTEND (pp,2*msize, RPTR(x),msize);

  MPZ_REALLOC_PTR (wp, w, msize);
  REDC_SMALL (wp, pp, m, MPTR(m), msize);
  wsize = msize;

  MPN_NORMALIZE (wp, wsize);
  SIZ(w) = wsize;
  TMP_FREE (marker);

  ASSERT_MPR_GET_Z (w, m);
}

static void
mpr_redc__get_z_big (mpz_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  mp_size_t  msize, wsize;
  mp_ptr     tp, qp, wp;
  mp_srcptr  mp;
  TMP_DECL (marker);

  TRACE (printf ("mpr_redc__get_z_big\n");
         mpn_trace ("  xp", RPTR(x), MRSIZE(m)));
  ASSERT_MPR_MOD (x, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  TMP_ALLOC_LIMBS_2 (tp, 2*msize, qp, 2*msize);

  mpn_mullowhalf_n (qp, RPTR(x), MINV(m), msize);
  TRACE (mpn_trace ("  qp", qp, msize));

  mp = MPTR(m);
  ASSERT_CODE (mpn_mul_n (tp, qp, MPTR(m), msize));
  ASSERT (mpn_cmp (tp, PTR(x), msize) == 0);

  mpn_mulhighhalf_n (tp, qp, mp, msize);
  TRACE (mpn_trace ("  sub", tp, 2*msize));

  MPZ_REALLOC (w, msize);
  wp = PTR(w);

  tp += msize;
  MPN_NONZERO_P_ACTION (tp, msize, goto nonzero);
  SIZ(w) = 0;
  goto done;

 nonzero:
  ASSERT_NOCARRY (mpn_sub_n (wp, mp, tp, msize));
  wsize = msize;
  TRACE (mpn_trace ("  diff", wp, wsize));

  MPN_NORMALIZE (wp, wsize);
  SIZ(w) = wsize;

 done:
  TMP_FREE (marker);

  ASSERT_MPR_GET_Z (w, m);
}

static int
mpr_redc__invert (mpr_ptr w, mpr_srcptr x, mpm_srcptr m)
{
  int  ret;
  ret = mpr_gen__invert (w, x, m);
  if (LIKELY (ret != 0))
    mpr_mul (w, w, m->_mp_r3mod, m);
  return ret;
}


/* (xR/m) = (x/m) if R is an even power of 2, which is always the case when
   GMP_NUMB_BITS is even, but if it's odd though we need to apply a factor
   (R/m).

   It'd be possible to apply (R/m) by munging the signs of the two operands,
   but that doesn't seem worth bothering with, mpz_jacobi still couldn't be
   a tail call.  */

#if GMP_NAIL_BITS == 0
#define mpr_redc__jacobi  mpr_gen__jacobi
#else
static int
mpr_redc__jacobi (mpr_srcptr x, mpm_srcptr m)
{
  mpz_t  xz;
  mp_srcptr  xp;
  mp_size_t  xsize, msize;
  mp_limb_t  mlow;
  int        r;

  ASSERT_MPR_MOD (x, m);

  xp = RPTR(x);
  msize = MRSIZE(m);
  xsize = MRSIZE(m);
  MPN_NORMALIZE (xp, xsize);

  PTR(xz) = (mp_ptr) xp;
  SIZ(xz) = xsize;
  ASSERT_CODE (ALLOC(xz) = MRSIZE(m));

  r = mpz_jacobi (xz, MZ(m));

  if (GMP_NUMB_BITS % 2 != 0 && r != 0)
    {
      mlow = MPTR(m)[0];
      r ^= ((msize & 1) << 1) & JACOBI_TWO_U_BIT1 (mlow);
      r = JACOBI_BIT1_TO_PN (r);
    }
  return r;
}
#endif

void
mpr_redc__mul_small (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     pp;
  TMP_DECL (marker);

  TRACE (printf ("mpr_redc__mul_small\n");
         mpn_trace ("  mul x", RPTR(x), MRSIZE(m));
         mpn_trace ("  mul y", RPTR(y), MRSIZE(m));
         mpn_trace ("  mod m", MPTR(m), MRSIZE(m)));
  ASSERT_MPR_MOD (x, m);
  ASSERT_MPR_MOD (y, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  pp = TMP_ALLOC_LIMBS (2*msize);

  mpn_mul_n (pp, RPTR(x), RPTR(y), msize);
  REDC_SMALL (RPTR(w), pp, m, MPTR(m), msize);

  ASSERT_MPR_MOD (w, m);
  TMP_FREE (marker);
}

void
mpr_redc__mul_big (mpr_ptr w, mpr_srcptr x, mpr_srcptr y, mpm_srcptr m)
{
  mp_size_t  msize;
  mp_ptr     pp, tp, qp;
  TMP_DECL (marker);

  TRACE (printf ("mpr_redc__mul_big\n");
         mpn_trace ("  mul x", RPTR(x), MRSIZE(m));
         mpn_trace ("  mul y", RPTR(y), MRSIZE(m)));
  ASSERT_MPR_MOD (x, m);
  ASSERT_MPR_MOD (y, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  TMP_ALLOC_LIMBS_3 (pp, 2*msize, tp, 2*msize, qp, 2*msize);

  mpn_mul_n (pp, RPTR(x), RPTR(y), msize);
  REDC_BIG (RPTR(w), pp, tp, qp, MPTR(m), MINV(m), msize);

  ASSERT_MPR_MOD (w, m);
  TMP_FREE (marker);
}

/* Enhancement: Sliding window, perhaps with on-demand calculation of
   intermediate powers.  */
static void
mpr_redc__pow_small (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  msize, esize, i;
  mp_ptr     rp, pp;
  mp_srcptr  bp, ep, mp;
  mp_limb_t  elimb, mask;
  int        lz;
  TMP_DECL (marker);

  ASSERT_MPR_MOD (b, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  rp = RPTR(r);

  esize = SIZ(e);
  if (esize == 0)
    {
      /* b^0 == 1 */
      MPN_COPY (rp, RPTR(MRMOD(m)), msize);
      return;
    }

  bp = RPTR(b);
  if (esize < 0)
    {
      esize = -esize;
      ASSERT_CARRY (mpr_redc__invert (r, b, m));
      bp = rp;
    }

  TMP_ALLOC_LIMBS_2SEP (pp,2*msize, rp,bp,msize);

  MPN_COPY (rp, bp, msize);  /* start b^1 */
  mp = MPTR(m);

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (lz, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - lz);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }
      TRACE (printf (" at i=%ld mask=%lX bit=%d\n",
                     i, mask, (elimb & mask) != 0));

      mpn_sqr_n (pp, rp, msize);
      REDC_SMALL (rp, pp, m, mp, msize);
      TRACE (mpn_trace (" double r", rp, msize));

      if ((elimb & mask) != 0)
        {
          mpn_mul_n (pp, rp, bp, msize);
          REDC_SMALL (rp, pp, m, mp, msize);
          TRACE (mpn_trace (" step   r", rp, msize));
        }
    }

  TMP_FREE (marker);
}

static void
mpr_redc__pow_big (mpr_ptr r, mpr_srcptr b, mpz_srcptr e, mpm_srcptr m)
{
  mp_size_t  msize, esize, i;
  mp_ptr     rp, pp, tp, qp;
  mp_srcptr  bp, ep, minv, mp;
  mp_limb_t  elimb, mask;
  int        lz;
  TMP_DECL (marker);

  ASSERT_MPR_MOD (b, m);

  TMP_MARK (marker);
  msize = MRSIZE(m);
  rp = RPTR(r);

  esize = SIZ(e);
  if (esize == 0)
    {
      /* b^0 == 1 */
      MPN_COPY (rp, RPTR(MRMOD(m)), msize);
      return;
    }

  bp = RPTR(b);
  if (esize < 0)
    {
      esize = -esize;
      ASSERT_CARRY (mpr_redc__invert (r, b, m));
      bp = rp;
    }

  TMP_ALLOC_LIMBS_4SEP (pp,2*msize, tp,2*msize, qp,2*msize, rp,bp,msize);

  MPN_COPY (rp, bp, msize);  /* start b^1 */
  mp = MPTR(m);
  minv = MINV(m);

  ep = PTR(e);
  i = esize-1;
  elimb = ep[i];
  count_leading_zeros (lz, elimb);
  mask = CNST_LIMB(1) << (BITS_PER_MP_LIMB - 1 - lz);

  for (;;)
    {
      mask >>= 1;
      if (mask == 0)
        {
          i--;
          if (i < 0)
            break;
          elimb = ep[i];
          mask = GMP_NUMB_HIGHBIT;
        }
      TRACE (printf (" at i=%ld mask=%lX bit=%d\n",
                     i, mask, (elimb & mask) != 0));

      mpn_sqr_n (pp, rp, msize);
      REDC_BIG (rp, pp, tp, qp, mp, minv, msize);
      TRACE (mpn_trace (" double r", rp, msize));

      if ((elimb & mask) != 0)
        {
          mpn_mul_n (pp, rp, bp, msize);
          REDC_BIG (rp, pp, tp, qp, mp, minv, msize);
          TRACE (mpn_trace (" step   r", rp, msize));
        }
    }

  ASSERT_MPR_MOD (r, m);
  TMP_FREE (marker);
}

static void
mpr_redc__set_si (mpr_ptr w, long x, mpm_srcptr m)
{
  mpr_mod__mul_si (w, m->_mp_rmod, x, m);
}

/* FIXME: Is there meant to be an mpn_tdiv_qr here?
   ENHANCE-ME: use division rather than mul by rmodm if z is smallish */
static void
mpr_redc__set_z (mpr_ptr w, mpz_srcptr x, mpm_srcptr m)
{
  mp_size_t  xsize = ABSIZ(x);
  if (0) /* BELOW_THRESHOLD (xsize, REDC_SETZ_THRESHOLD)) */
    {
      mpz_t      xr;
      mp_ptr     xrp;
      mp_size_t  xrsize, msize;
      TMP_DECL (marker);
      TMP_MARK (marker);
      msize = MRSIZE(m);
      xrsize = xsize + msize;
      xrp = TMP_ALLOC_LIMBS (xrsize);
      MPN_ZERO (xrp, msize);
      MPN_COPY (xrp+msize, PTR(x), xsize);
      PTR(xr) = xrp;
      SIZ(xr) = (SIZ(x) >= 0 ? xrsize : -xrsize);
      ASSERT_CODE (ALLOC(xr) = xsize + msize);
      TMP_FREE (marker);
    }
  else
    {
      mpr_mod__set_z (w, x, m);
      mpr_mul (w, w, m->_mp_r2mod, m);
    }
  ASSERT_MPR_MOD (w, m);
}

static void
mpr_redc__set_ui (mpr_ptr w, unsigned long x, mpm_srcptr m)
{
  if (UNLIKELY (x > GMP_NUMB_MAX))
    {
      /* FIXME: Think this idea was only to avoid two-limb division in
         mpr_mod__mul_ui, but maybe that ought to be ok anyway.  */
      MPZ_FAKE_UI_VAR (xz, x);
      mpr_redc__set_z (w, xz, m);
      return;
    }

  mpr_mod__mul_ui (w, m->_mp_rmod, x, m);
}

/* Same as mpr_redc__add_ui, but using MNEGONE(m). */
static void
mpr_redc__sub_ui (mpr_ptr w, mpr_srcptr x, unsigned long y, mpm_srcptr m)
{
  mp_ptr     wp;
  mp_srcptr  xp;
  mp_size_t  msize;
  mp_limb_t  c;

  ASSERT_MPR_MOD (x, m);

  if (UNLIKELY (y > GMP_NUMB_MAX))
    {
      mpr_t  yr;
      mpr_init_set_ui (yr, y, m);
      mpr_sub (w, x, yr, m);
      mpr_clear (yr, m);
      return;
    }

  wp = RPTR(w);
  xp = RPTR(x);
  msize = MRSIZE(m);
  if (w != x)
    MPN_COPY (wp, xp, msize);
  c = mpn_addmul_1 (wp, RPTR(MNEGONE(m)), msize, (mp_limb_t) y);
  mpr_mod__reduce_1 (c, wp, m);

  ASSERT_MPR_MOD (w, m);
}

static void
mpr_redc__mpm_clear (mpm_ptr m)
{
  mp_size_t  msize;

  msize = MRSIZE(m);
  __GMP_FREE_FUNC_LIMBS (MRMOD(m), msize);
  __GMP_FREE_FUNC_LIMBS (MR2MOD(m), msize);
  __GMP_FREE_FUNC_LIMBS (MR3MOD(m), msize);
  __GMP_FREE_FUNC_LIMBS (MNEGONE(m), msize);

  if (MINV(m) != NULL)
    __GMP_FREE_FUNC_LIMBS (MINV(m), msize);

  mpm_mod__clear (m);
}

void
mpm_init_redc (mpm_ptr m, mpz_srcptr mz)
{
  mp_size_t  msize;
  mp_ptr     mp, mrmod, mr2mod, mr3mod, minv;
  mpz_t      rz, pz;

  msize = SIZ(mz);
  if (msize == 0)
    DIVIDE_BY_ZERO;

  msize = ABS(msize);
  if (msize == 1)
    {
      mpm_init_one (m, PTR(mz)[0]);
      return;
    }

  mpm_mod__init (m, mz);
  msize = MRSIZE(m);

  m->_mpm_clear = mpr_redc__mpm_clear;

  m->_mpr_add          = mpr_mod__add;
  m->_mpr_add_ui       = mpr_redc__add_ui;
  m->_mpr_addmul_ui    = mpr_mod__addmul_ui;
  m->_mpr_clear        = mpr_gen__clear;
  m->_mpr_div          = mpr_gen__div__invert_mul;
  m->_mpr_div_2exp     = mpr_mod__div_2exp;
  m->_mpr_equal        = mpr_gen__equal;
  m->_mpr_equal_one    = mpr_redc__equal_one;
  m->_mpr_equal_negone = mpr_gen__equal_negone;
  m->_mpr_equal_ui     = mpr_redc__equal_ui;
  m->_mpr_equal_zero   = mpr_gen__equal_zero;
  m->_mpr_gcd          = mpr_gen__gcd__odd;  /* factor R ok since m odd */

  /* It doesn't seem possible to do just some of the redc in mpr_redc__get_z,
     since the high limbs are needed to know whether to add (or subtract) a
     copy of m in the final result.  */
  m->_mpr_get_ui       = mpr_gen__get_ui__get_z;

  m->_mpr_get_z        = (BELOW_THRESHOLD (msize, REDC_BIG_THRESHOLD)
                          ? mpr_redc__get_z_small : mpr_redc__get_z_big);
  m->_mpr_invert       = mpr_redc__invert;
  m->_mpr_jacobi       = mpr_redc__jacobi;
  m->_mpr_mul          = (BELOW_THRESHOLD (msize, REDC_BIG_THRESHOLD)
                          ? mpr_redc__mul_small : mpr_redc__mul_big);
  m->_mpr_mul_2exp     = mpr_mod__mul_2exp;
  m->_mpr_mul_si       = mpr_mod__mul_si;
  m->_mpr_mul_ui       = mpr_mod__mul_ui;
  m->_mpr_neg          = mpr_mod__neg;
  m->_mpr_new          = mpr_gen__new;
  m->_mpr_pow          = (BELOW_THRESHOLD (msize, REDC_BIG_THRESHOLD)
                          ? mpr_redc__pow_small : mpr_redc__pow_big);
  m->_mpr_set          = mpr_gen__set;
  m->_mpr_set_z        = mpr_redc__set_z;
  m->_mpr_set_si       = mpr_redc__set_si;
  m->_mpr_set_ui       = mpr_redc__set_ui;
  m->_mpr_sub          = mpr_mod__sub;
  m->_mpr_sub_ui       = mpr_redc__sub_ui;
  m->_mpr_submul_ui    = mpr_mod__submul_ui;
  m->_mpr_swap         = mpr_gen__swap;
  m->_mpr_urandomm     = mpr_mod__urandomm;

  mp = MPTR(m);

  if (msize <= 0 || (mp[0] & 1) == 0)
    DIVIDE_BY_ZERO;

  mpz_init_set_ui (rz, 1L);
  mpz_mul_2exp (rz, rz, msize*GMP_NUMB_BITS);
  mpz_tdiv_r (rz, rz, MZ(m));

  /* R mod M */
  mrmod = __GMP_ALLOCATE_FUNC_LIMBS (msize);
  MRMOD(m) = (mpr_t) mrmod;
  MPN_COPY_ZERO_EXTEND (mrmod, msize, PTR(rz), SIZ(rz));
  TRACE (mpn_trace ("  rmod", MRMOD(m), msize));

  /* -1 as a residue, is -R mod M */
  {
    mp_ptr  mnegone;
    mnegone = __GMP_ALLOCATE_FUNC_LIMBS (msize);
    MNEGONE(m) = (mpr_ptr) mnegone;
    ASSERT_NOCARRY (mpn_sub_n (mnegone, mp, mrmod, msize));
    ASSERT_MPR_MOD (m->_mp_negone, m);
  }

  /* R^2 mod M */
  mpz_init (pz);
  mpz_mul (pz, rz, rz);
  mpz_tdiv_r (pz, pz, MZ(m));
  mr2mod = __GMP_ALLOCATE_FUNC_LIMBS (msize);
  MR2MOD(m) = (mpr_ptr) mr2mod;
  MPN_COPY_ZERO_EXTEND (mr2mod, msize, PTR(pz), SIZ(pz));
  TRACE (mpn_trace ("  r2mod", MR2MOD(m), msize));

  /* R^3 mod M */
  mpz_mul (pz, pz, rz);
  mpz_tdiv_r (pz, pz, MZ(m));
  mr3mod = __GMP_ALLOCATE_FUNC_LIMBS (msize);
  MR3MOD(m) = (mpr_ptr) mr3mod;
  MPN_COPY_ZERO_EXTEND (mr3mod, msize, PTR(pz), SIZ(pz));
  TRACE (mpn_trace ("  r3mod", MR3MOD(m), msize));

  if (MSMALLINV_SIZE == 1)
    {
      modlimb_neginvert (MSMALLINV(m)[0], mp[0]);
    }
  else if (MSMALLINV_SIZE == 2)
    {
      mp_limb_t inv1, inv0;
      modlimb_invert_2 (inv1, inv0, (msize > 1 ? mp[1] : 0), mp[0]);
      NEG_2_LOWNZ (MSMALLINV(m)[1], MSMALLINV(m)[0], inv0, inv1);
    }
  else
    {
      /* ENHANCE-ME: Should be able to do the 3 or 4 limb cases inline. */
      mpn_modlimbs_invert (MSMALLINV(m), mp, MSMALLINV_SIZE);
      MPN_NEG_N_NONZERO (MSMALLINV(m), MSMALLINV(m), MSMALLINV_SIZE);
    }

  minv = NULL;
  if (ABOVE_THRESHOLD (msize, REDC_BIG_THRESHOLD))
    {
      minv = __GMP_ALLOCATE_FUNC_LIMBS (msize);
      mpn_modlimbs_invert (minv, mp, msize);
    }
  MINV(m) = minv;

  mpz_clear (rz);
  mpz_clear (pz);
}
